@@ -1,4 +1,5 @@
 #include <string.h>
+#include <stdlib.h>
 #include "render.h"
 #include "system.h"
 #include "io.h"
@@ -357,21 +358,30 @@ void handle_binding_up(keybinding * binding)
 				if (!screenshot_base) {
 					screenshot_base = "$HOME";
 				}
+				const system_media *media = current_media();
 				tern_node *vars = tern_insert_ptr(NULL, "HOME", get_home_dir());
 				vars = tern_insert_ptr(vars, "EXEDIR", get_exe_dir());
+				vars = tern_insert_ptr(vars, "USERDATA", (char *)get_userdata_dir());
+				vars = tern_insert_ptr(vars, "ROMNAME", media->name);
+				vars = tern_insert_ptr(vars, "ROMDIR", media->dir);
 				screenshot_base = replace_vars(screenshot_base, vars, 1);
 				tern_free(vars);
+				ensure_dir_exists(screenshot_base);
 				time_t now = time(NULL);
 				struct tm local_store;
 				char fname_part[256];
 				char *template = tern_find_path(config, "ui\0screenshot_template\0", TVAL_PTR).ptrval;
-				if (!template) {
-					template = "blastem_%c.ppm";
+				if (template) {
+					vars = tern_insert_ptr(NULL, "ROMNAME", media->name);
+					template = replace_vars(template, vars, 0);
+				} else {
+					template = strdup("blastem_%c.ppm");
 				}
 				strftime(fname_part, sizeof(fname_part), template, localtime_r(&now, &local_store));
 				char const *parts[] = {screenshot_base, PATH_SEP, fname_part};
 				char *path = alloc_concat_m(3, parts);
 				free(screenshot_base);
+				free(template);
 				render_save_screenshot(path);
 			}
 			break;
@@ -557,7 +567,7 @@ int parse_binding_target(int device_num, char * target, tern_node * padbuttons,
 {
 	const int gpadslen = strlen("gamepads.");
 	const int mouselen = strlen("mouse.");
-	if (!strncmp(target, "gamepads.", gpadslen)) {
+	if (startswith(target, "gamepads.")) {
 		int padnum = target[gpadslen] == 'n' ? device_num + 1 : target[gpadslen] - '0';
 		if (padnum >= 1 && padnum <= 8) {
 			int button = tern_find_int(padbuttons, target + gpadslen + 1, 0);
@@ -575,7 +585,7 @@ int parse_binding_target(int device_num, char * target, tern_node * padbuttons,
 		} else {
 			warning("Gamepad mapping string '%s' refers to an invalid gamepad number %c\n", target, target[gpadslen]);
 		}
-	} else if(!strncmp(target, "mouse.", mouselen)) {
+	} else if(startswith(target, "mouse.")) {
 		int mousenum = target[mouselen] == 'n' ? device_num + 1 : target[mouselen] - '0';
 		if (mousenum >= 1 && mousenum <= 8) {
 			int button = tern_find_int(mousebuttons, target + mouselen + 1, 0);
@@ -593,7 +603,7 @@ int parse_binding_target(int device_num, char * target, tern_node * padbuttons,
 		} else {
 			warning("Gamepad mapping string '%s' refers to an invalid mouse number %c\n", target, target[mouselen]);
 		}
-	} else if(!strncmp(target, "ui.", strlen("ui."))) {
+	} else if(startswith(target, "ui.")) {
 		if (!strcmp(target + 3, "vdp_debug_mode")) {
 			*subtype_a = UI_DEBUG_MODE_INC;
 		} else if(!strcmp(target + 3, "vdp_debug_pal")) {
@@ -603,7 +613,7 @@ int parse_binding_target(int device_num, char * target, tern_node * padbuttons,
 			*subtype_a = UI_ENTER_DEBUGGER;
 		} else if(!strcmp(target + 3, "save_state")) {
 			*subtype_a = UI_SAVE_STATE;
-		} else if(!strncmp(target + 3, "set_speed.", strlen("set_speed."))) {
+		} else if(startswith(target + 3, "set_speed.")) {
 			*subtype_a = UI_SET_SPEED;
 			*subtype_b = atoi(target + 3 + strlen("set_speed."));
 		} else if(!strcmp(target + 3, "next_speed")) {
@@ -815,7 +825,7 @@ void process_pad_button(char *key, tern_val val, uint8_t valtype, void *data)
 			if (hostbutton == RENDER_INVALID_NAME) {
 				warning("%s is not a valid gamepad input name\n", key);
 			} else if (hostbutton == RENDER_NOT_MAPPED && hostpadnum != map_warning_pad) {
-				warning("No SDL 2 mapping exists for input %s on gamepad %d\n", key, hostpadnum);
+				debug_message("No SDL 2 mapping exists for input %s on gamepad %d\n", key, hostpadnum);
 				map_warning_pad = hostpadnum;
 			}
 			return;
@@ -824,7 +834,7 @@ void process_pad_button(char *key, tern_val val, uint8_t valtype, void *data)
 			bind_dpad(hostpadnum, render_dpad_part(hostbutton), render_direction_part(hostbutton), bindtype, subtype_a, subtype_b);
 			return;
 		} else if (hostbutton & RENDER_AXIS_BIT) {
-			bind_axis(hostpadnum, render_axis_part(hostbutton), 1, bindtype, subtype_a, subtype_b);
+			bind_axis(hostpadnum, render_axis_part(hostbutton), hostbutton & RENDER_AXIS_POS, bindtype, subtype_a, subtype_b);
 			return;
 		}
 	}
@@ -862,7 +872,7 @@ void process_pad_axis(char *key, tern_val val, uint8_t valtype, void *data)
 			if (axis == RENDER_INVALID_NAME) {
 				warning("%s is not a valid gamepad input name\n", key);
 			} else if (axis == RENDER_NOT_MAPPED && hostpadnum != map_warning_pad) {
-				warning("No SDL 2 mapping exists for input %s on gamepad %d\n", key, hostpadnum);
+				debug_message("No SDL 2 mapping exists for input %s on gamepad %d\n", key, hostpadnum);
 				map_warning_pad = hostpadnum;
 			}
 			goto done;
