@@ -3,7 +3,11 @@
 
 #include <stdint.h>
 #include "m68k_core.h"
+#ifdef NEW_CORE
+#include "z80.h"
+#else
 #include "z80_to_x86.h"
+#endif
 
 typedef struct disp_def {
 	struct disp_def * next;
