@@ -80,7 +80,7 @@ char *replace_vars(char *base, tern_node *vars, uint8_t allow_env)
 	for (char *cur = base; *cur; ++cur)
 	{
 		if (in_var) {
-			if (!(*cur == '_' || isalnum(*cur))) {
+			if (!isalnum(*cur)) {
 				positions[num_vars].end = cur-base;
 				if (positions[num_vars].end - positions[num_vars].start > max_var_len) {
 					max_var_len = positions[num_vars].end - positions[num_vars].start;
@@ -189,6 +189,11 @@ char * split_keyval(char * text)
 	return text+1;
 }
 
+uint8_t startswith(const char *haystack, const char *prefix)
+{
+	return !strncmp(haystack, prefix, strlen(prefix));
+}
+
 void bin_to_hex(uint8_t *output, uint8_t *input, uint64_t size)
 {
 	while (size)
@@ -472,13 +477,16 @@ void warning(char *format, ...)
 	va_end(args);
 }
 
+static uint8_t output_enabled = 1;
 void info_message(char *format, ...)
 {
 	va_list args;
 	va_start(args, format);
 #ifndef _WIN32
 	if (headless || (isatty(STDOUT_FILENO) && isatty(STDIN_FILENO))) {
-		info_printf(format, args);
+		if (output_enabled) {
+			info_printf(format, args);
+		}
 	} else {
 #endif
 		int32_t size = strlen(format) * 2;
@@ -498,7 +506,9 @@ void info_message(char *format, ...)
 			va_start(args, format);
 			vsnprintf(buf, actual, format, args);
 		}
-		info_puts(buf);
+		if (output_enabled) {
+			info_puts(buf);
+		}
 		render_infobox("BlastEm Info", buf);
 		free(buf);
 #ifndef _WIN32
@@ -507,6 +517,20 @@ void info_message(char *format, ...)
 	va_end(args);
 }
 
+void debug_message(char *format, ...)
+{
+	va_list args;
+	va_start(args, format);
+	if (output_enabled) {
+		info_printf(format, args);
+	}
+}
+
+void disable_stdout_messages(void)
+{
+	output_enabled = 0;
+}
+
 #ifdef _WIN32
 #include <windows.h>
 #include <shlobj.h>
@@ -830,9 +854,20 @@ void sort_dir_list(dir_entry *list, size_t num_entries)
 	qsort(list, num_entries, sizeof(dir_entry), sort_dir_alpha);
 }
 
+uint8_t delete_file(char *path)
+{
+#ifdef _WIN32
+	//TODO: Call Unicode version and prepend special string to remove max path limitation
+	return 0 != DeleteFileA(path);
+#else
+	return 0 == unlink(path);
+#endif
+}
+
 #ifdef __ANDROID__
 
 #include <SDL.h>
+#ifndef IS_LIB
 char *read_bundled_file(char *name, uint32_t *sizeret)
 {
 	SDL_RWops *rw = SDL_RWFromFile(name, "rb");
@@ -860,6 +895,7 @@ char *read_bundled_file(char *name, uint32_t *sizeret)
 	SDL_RWclose(rw);
 	return ret;
 }
+#endif
 
 char const *get_config_dir()
 {
@@ -873,6 +909,7 @@ char const *get_userdata_dir()
 
 #else
 
+#ifndef IS_LIB
 char *read_bundled_file(char *name, uint32_t *sizeret)
 {
 #ifdef DATA_PATH
@@ -916,7 +953,7 @@ char *read_bundled_file(char *name, uint32_t *sizeret)
 	fclose(f);
 	return ret;
 }
-
+#endif //ISLIB
 
 #ifdef _WIN32
 char const *get_userdata_dir()
@@ -973,8 +1010,6 @@ char const *get_userdata_dir()
 }
 
 
-#endif
+#endif //_WIN32
+#endif //__ANDROID__
 
-
-
-#endif
